import AlteryxPythonSDK as Sdk
from typing import IO, Union

import azure.datalake.store.core as adl_store
from adal import AdalError
from azure.datalake.store.exceptions import DatalakeRESTException

import connectors_core.alteryx as alteryx
from connectors_core.azure import AzureAuthenticationType, AdlSpAuthentication, AdlEndUserAuthentication
from connectors_core.file import DelimitedReader, JsonPivotInputReader, AvroInputReader, FileFormat


class AyxPlugin(alteryx.AyxPluginBase):
    """
    Implements the plugin interface methods, to be utilized by the Alteryx engine to communicate with this plugin.
    Prefixed with "pi", the Alteryx engine will expect the below five interface methods to be defined.
    """

    def __init__(self, n_tool_id: int, alteryx_engine: Sdk.AlteryxEngine, output_anchor_mgr: Sdk.OutputAnchorManager):
        """
        Constructor is called whenever the Alteryx engine wants to instantiate an instance of this plugin.
        :param n_tool_id: The assigned unique identification for a tool instance.
        :param alteryx_engine: Provides an interface into the Alteryx engine.
        :param output_anchor_mgr: A helper that wraps the outgoing connections for a plugin.
        """
        super().__init__(n_tool_id, alteryx_engine, output_anchor_mgr)

        # Configuration Options
        self.adl_auth_type = AzureAuthenticationType.ServicePrincipal
        self.adl_store_name = None  # type: str
        self.adl_tenant_id = None  # type: str
        self.adl_client_id = None  # type: str
        self.adl_secret_key = None  # type: str
        self.adl_username = None  # type: str
        self.adl_password = None  # type: str
        self.adl_file_path = ''
        self.adl_file_format = FileFormat.Delimited
        self.adl_delimited_has_header = True
        self.adl_field_size = 255
        self.adl_field_delimiter = ','
        self.adl_field_quote_character = '"'
        self.adl_encoding = 'utf-8'
        self.adl_json_force_string_values = True
        self.adl_start_import_at_line = 1

        self.adl_authentication = None

    def pi_init(self, str_xml: str):
        """
        Handles extracting user-entered file name, and input data verification.
        Called when the Alteryx engine is ready to provide the tool configuration from the GUI.
        :param str_xml: The raw XML from the GUI.
        """

        # this will set is_initialized=True, which will be set to False if the config is invalid
        super().pi_init(str_xml)

        # Getting the user-entered file name string from the GUI, and the output anchor from the XML file.
        self.adl_store_name = self.get_xml_config_value(str_xml, 'adl_StoreName')
        self.adl_auth_type = AzureAuthenticationType[
            self.get_xml_config_value(str_xml, 'adl_AuthenticationType')
            or AzureAuthenticationType.Unknown.name
            ]
        self.adl_tenant_id = self.get_xml_config_value(str_xml, 'adl_TenantId') or ''

        # Validation
        if not self.adl_store_name:
            self.invalidate_config(
                "Please provide your Azure Data Lake Store Name.")
        if self.adl_auth_type == AzureAuthenticationType.Unknown:
            self.invalidate_config("Please specify an Authentication Type")
        if not self.adl_tenant_id:
            self.invalidate_config("Please provide a Tenant Id")

        if self.adl_auth_type == AzureAuthenticationType.ServicePrincipal:
            self.adl_client_id = self.get_xml_config_value(str_xml, 'adl_ClientId') or ''
            self.adl_secret_key = self.get_encrypted_xml_config_value(str_xml, 'adl_SecretKey') or ''

            # Validation
            if not self.adl_client_id:
                self.invalidate_config("Please provide a Client Id")
            if not self.adl_secret_key:
                self.invalidate_config("Please provide a Secret Key")

            if self.is_initialized:
                self.adl_authentication = AdlSpAuthentication(
                    store_name=self.adl_store_name,
                    tenant_id=self.adl_tenant_id,
                    client_id=self.adl_client_id,
                    secret_key=self.adl_secret_key
                )
        elif self.adl_auth_type == AzureAuthenticationType.EndUser:
            self.adl_username = self.get_xml_config_value(str_xml, 'adl_UserName')
            self.adl_password = self.get_encrypted_xml_config_value(str_xml, 'adl_Password')

            if not self.adl_username:
                self.invalidate_config("Please provide a User Name")
            if not self.adl_password:
                self.invalidate_config("Please provide a Password")

            if self.is_initialized:
                self.adl_authentication = AdlEndUserAuthentication(
                    tenant_id=self.adl_tenant_id,
                    username=self.adl_username,
                    password=self.adl_password)

        # File Configuration
        self.adl_file_format = FileFormat[
            self.get_xml_config_value(str_xml, 'adl_FileFormat') or FileFormat.Unknown.name
            ]
        self.adl_file_path = self.get_xml_config_value(str_xml, 'adl_FilePath')
        self.adl_json_force_string_values = (self.get_xml_config_value(str_xml, 'adl_JSONForceStringValues') == "True")

        self.adl_delimited_has_header = (self.get_xml_config_value(str_xml, 'adl_DelimitedHasHeader') == "True")
        self.adl_start_import_at_line = int(self.get_xml_config_value(str_xml, 'adl_StartImportAtLine') or 1)

        try:
            if self.adl_start_import_at_line < 1:
                self.invalidate_config("Invalid Starting Line: must be greater or equal to 1")
        except ValueError:
            self.invalidate_config("Invalid Starting Line: must be a number")

        try:
            self.adl_field_size = int(self.get_xml_config_value(str_xml, 'adl_MaxFieldLength') or 255)
            if self.adl_field_size < 1:
                self.invalidate_config("Invalid Max Field Length: must be greater than zero")
        except ValueError:
            self.invalidate_config("Invalid Max Field Length: must be a number")

        try:
            self.adl_field_quote_character = {
                'QUOTE': '\'',
                'DOUBLE_QUOTE': '"',
                'BACKTICK': '`'
            }[self.get_xml_config_value(str_xml, 'adl_QuoteCharacter') or 'DOUBLE_QUOTE']
        except KeyError:
            self.invalidate_config("Invalid Quote Character")

        self.adl_encoding = self.get_xml_config_value(str_xml, 'adl_CodePage') or 'utf-8'

        try:
            self.adl_field_delimiter = {
                'COMMA': ',',
                'TAB': '\t',
                'SPACE': ' ',
                'PIPE': '|'
            }[self.get_xml_config_value(str_xml, 'adl_Delimiter') or 'COMMA']
        except KeyError:
            self.invalidate_config("Invalid Field Delimiter")

        if not self.adl_file_path:
            # TODO: enforce starting file paths with / ?
            self.invalidate_config("Please provide a valid File Path")
        if self.adl_file_format == FileFormat.Unknown:
            self.invalidate_config("Please specify a File Format")

        self.output_anchor = self.output_anchor_mgr.get_output_anchor('Output')  # type: Sdk.OutputAnchor

    def pi_add_incoming_connection(self, str_type: str, str_name: str) -> Union[alteryx.IncomingConnectionBase, None]:
        """
        The IncomingInterface objects are instantiated here, one object per incoming connection, however since
        this tool does not accept an incoming connection, instantiation is not needed and "ii" methods won't be called.
        Called when the Alteryx engine is attempting to add an incoming data connection.
        :param str_type: The name of the input connection anchor, defined in the Config.xml file.
        :param str_name: The name of the wire, defined by the workflow author.
        :return: self.
        """

        self.display_error_msg("This tool does not accept incoming connections")

        return None

    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        """
        Called when the Alteryx engine is attempting to add an outgoing data connection.
        :param str_name: The name of the output connection anchor, defined in the Config.xml file.
        :return: True signifies that the connection is accepted.
        """

        # self.display_info_msg("Added connection {}".format(str_name))

        return True

    def pi_push_all_records(self, n_record_limit: int = -1) -> bool:
        """
        Handles pushing records out to downstream tool(s).
        Called when a tool has no incoming data connection.
        :param n_record_limit: Set it to <0 for no limit, 0 for no records, and >0 to specify the number of records.
        :return: False if there are issues with the input data or if the workflow isn't being ran, otherwise True.
        """

        if not self.is_initialized:
            return False

        update_only = self.alteryx_engine.get_init_var(self.n_tool_id, 'UpdateOnly') == 'True' \
                      or n_record_limit == 0

        try:
            adl_client = self.get_adl_client()
        except AdalError as e:
            self.display_error_msg("There was a problem authenticating with Azure Data Lake: {}".format(e))
            return False
        try:
            file_meta = adl_client.info(path=self.adl_file_path)
            # See here for file info properties:
            # http://hadoop.apache.org/docs/stable/hadoop-project-dist/hadoop-hdfs/WebHDFS.html#Status_of_a_FileDirectory
            file_length_bytes = int(file_meta['length'])

            with adl_client.open(path=self.adl_file_path, mode='rb') as input_stream:
                if self.adl_file_format == FileFormat.Delimited:
                    reader = DelimitedReader(
                        input_stream,
                        self.output_anchor,
                        self.alteryx_engine,
                        file_length_bytes,
                        self.get_field_source,
                        binary_input=True,
                        max_field_size=self.adl_field_size,
                        delimiter=self.adl_field_delimiter,
                        quotechar=self.adl_field_quote_character,
                        encoding=self.adl_encoding,
                        has_header_row=self.adl_delimited_has_header,
                        start_data_at_line=self.adl_start_import_at_line
                    )

                elif self.adl_file_format == FileFormat.Avro:
                    reader = AvroInputReader(
                        input_stream,
                        self.output_anchor,
                        self.alteryx_engine,
                        file_length_bytes,
                        self.get_field_source)

                elif self.adl_file_format == FileFormat.JSON:
                    reader = JsonPivotInputReader(
                        input_stream,
                        self.output_anchor,
                        self.alteryx_engine,
                        file_length_bytes,
                        self.get_field_source,
                        force_string_values=self.adl_json_force_string_values,
                        max_field_size=self.adl_field_size)

                else:
                    self.display_error_msg("Unknown File Format!")
                    return False

                reader.init()
                if update_only:
                    return True

                record_count = reader.read(n_record_limit)

            self.display_info_msg(
                "{0} records were read from {1}/{2}".format(record_count, self.adl_store_name, self.adl_file_path)
            )
            self.output_anchor.update_progress(1.0)
            self.output_anchor.close()  # Close outgoing connections.
            return True
        except FileNotFoundError as e:
            self.display_error_msg("Unable to open file: {}".format(e))
            return False
        except ConnectionError as e:
            self.display_error_msg("Error establishing connection: {}".format(e))
            return False
        except DatalakeRESTException as e:
            self.display_error_msg("Error establishing connection: {}".format(e))
            return False

    def pi_close(self, b_has_errors: bool):
        super().pi_close(b_has_errors)
        self.output_anchor.assert_close()  # Checks whether connections were properly closed.

    def report_input_stream_progress(self, stream: IO, stream_length):
        stream_progress = round(stream.tell() / stream_length, 2)
        if (stream_progress * 100) % 5 == 0:
            self.output_anchor.update_progress(stream_progress)

    def get_adl_client(self) -> adl_store.AzureDLFileSystem:
        """
        Initialize a new ADL File System Client
        :return: an Azure Data Lake File System Client
        """
        if not self.is_initialized:
            self.display_error_msg("Input has not been initialized")
            raise Exception("Not Initialized")

        credentials = self.adl_authentication.get_credentials()
        return adl_store.AzureDLFileSystem(store_name=self.adl_store_name, token=credentials)

    def get_field_source(self, field: Sdk.Field):
        return "ADLS: https://{0}.azuredatalakestore.net{1}".format(
            self.adl_store_name,
            self.adl_file_path
        )
